#=

Program for finding extreme Delacorte Numbers
Copyright (c) 2015, Arch D. Robison
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors
   may be used to endorse or promote products derived from this software without
   specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.

=#

if !isdefined(:N)
    const N = 27
end

println("N=",N)

# pinned contains values that should not be moved.
if !isdefined(:pinned)
    pinned = Int[]
else
    println("pinned = ",pinned)
end

import Base.copy, Base.getindex, Base.isless, Base.length, Base.rand, Base.show, Base.size, Base.string
importall Base.Order

# Nummeric type used for Delacorte calculations.
# Should be signed type big enough to hold a coordinate or 1st moment.
# Int32 or Float32 are good choices.
const QType = Int32

# Define fast conversions to Int32.
fast_int32(x::Integer) = x % Int32
fast_int32(x::FloatingPoint) = Base.unsafe_trunc(Int32,x)

# ---------- Random number support ---------

# Holds values for stream of random numbers in 0:n-1
immutable RandStream
    modulus::Uint32
    limit::Uint32

    # Construct stream for random numbers in 0:n-1
    function RandStream( n::Integer )
        @assert 1<=n
        @assert n<=2^24
        k = convert(Uint32,n)
        z = zero(Uint32)
        m = (z-k) % k
        new( k, m==0 ? z-k : z-m )
    end
end

# Get a random number from stream s
function rand( s::RandStream )
    r = rand(Uint32)
    while r>=s.limit
        r = rand(Uint32)
    end
    int(r % s.modulus)
end

# ---------- Lookup tables for weight between two numbers ---------

# Find last non-zero element
function findlast(a)
    k = length(a)
    while k>0 && a[k]==0
        k -= 1
    end
    k
end

function makeWeight()
    n = N^2
    t = QType[i!=j ? gcd(i,j)-1 : 0 for i=1:n, j=1:n]
    s = QType[sum(t[:,i]) for i=1:n]
    p = sortperm(s,alg=MergeSort,rev=true)
    permute!(s,p)
    m = findlast(s)
    w = QType[t[p[i],p[j]] for i=1:m, j=1:n]
    (w,s,p)
end

# These three tables are indexed by encoded values.
const (weight,weightSum,decode) = makeWeight()

# Table of encodings
const encode = invperm(decode)

println("weight built")

# ---------- sumOfPairDistances ---------

# Compute sum of all dist2
function makeSumOfPairDistances()
    s = 0
    for x0=1:N, x1=1:N
        s += (x1-x0)^2
    end
    s*N^2
end

# Sum of all pair distances
const sumOfPairDistances = makeSumOfPairDistances()

# Slow Delacorte calculation for checking
function slowDelacorte{T<:Integer}( a::Matrix{T} )
    (m,n) = size(a)
    s = 0
    for x0=1:n, x1=1:n, y0=1:m, y1=1:m
        s += gcd(a[y0,x0],a[y1,x1]) * ((x0-x1)^2 + (y0-y1)^2)
    end
    @assert iseven(s)
    div(s,2)
end

# ---------- Operations on abstract Delacorte Rectangle ----------

abstract Delacorte

# Return matrix representation of Delacorte transformed by function f.
function rect(f::Function, r::Delacorte)
    (m,n) = size(r)
    a = zeros(Int,m,n)
    for k=1:m*n
        (j,i) = r[k]
        @assert a[i,j]==0
        a[i,j] = f(k)
    end
    a
end

# Convert abstract representation to matrix representation
rect( r::Delacorte ) = rect(k->decode[k],r)

# Get rectangle of corresponding weightsum values.
wsmask( r::Delacorte ) = rect(k->int(weightSum[k]),r)

# Print array in AZPC format.
function printAZPC{T}( f::IO, a::Matrix{T} )
    (m,n) = size(a)
    ndig = trunc(Int,log10(m*n))+1
    for i = 1:m
        print( f, "(")
        for j = 1:n
            j==1 || print( f, ", ")
            print( f, lpad(string(a[i,j]),ndig," ") )
        end
        print( f, ")")
        i==m || println( f, "," )
    end
end

# Show Delacorte rectangle
function show( f::IO, r::Delacorte )
    println(f,"Size = $N x $N")
    d = delacorte(r)
    a = rect(r)
    e = slowDelacorte(a)
    if e==d
        println(f,"Delacorte# = ",d)
    else
        println("ERROR: slow Delacorte# = ",e," fast Delacorte# = ",d)
    end
    printAZPC(f,a)
end

# ---------- Moment Rep of a Delacorte Rectangle ----------

type Quad <: Delacorte
    z :: Matrix{QType}     # Coordinates of points
    c :: Matrix{QType}     # First moments points
    d :: Int               # Delacorte number
    function Quad( z::Matrix{QType} )
        c = computeMoments(z)
        new(z, c, computeDelacorte(z,c))
    end
end

# Compute Delacorte "first moments" of points in z
function computeMoments{T<:Real}( z::Matrix{T} )
    w = weight
    (m,n) = size(w)
    @assert m≤n
    @assert size(z)==(n,2)
    c = zeros(T,n,2)
    @inbounds for i=1:m
        Σx = zero(T)
        Σy = zero(T)
        @simd for j=1:m
            Σx += w[j,i]*z[j,1]
            Σy += w[j,i]*z[j,2]
        end
        c[i,1] = Σx
        c[i,2] = Σy
    end
    c
end

function computeDelacorte{T<:Real,U<:Real}( z::Matrix{T}, c::Matrix{U} )
    (m,n) = size(weight)
    s = 0
    @inbounds @simd for i=1:m
        s += fast_int32(weightSum[i]*(z[i,1]^2+z[i,2]^2) - (c[i,1]*z[i,1] + c[i,2]*z[i,2]))
    end
    s + sumOfPairDistances
end

# Verify that q is self-consistent
function verify( q::Quad )
    @assert q.c==computeMoments(q.z)
    @assert q.d==slowDelacorte(rect(q))
    nothing
end

delacorte( q::Quad ) = q.d

# Number of points in the Quad
length( q::Quad ) = size(q.z,1)

# Get index of a given point
getindex( q::Quad, i::Integer ) = (q.z[i,1],q.z[i,2])

# Get deep copy of q
copy( q::Quad ) = deepcopy(q)

# Size of rect(q)
size( q::Quad ) = (int(maximum(q.z[:,2])),int(maximum(q.z[:,1])))

# Order for sorting Quads by Delacorte value
isless( q::Quad, r::Quad ) = q.d<r.d

# Compute change in delacorte(q) if numbers with encodings i and j are swapped
function delta( q::Quad, i::Integer, j::Integer )
    (m,n) = size(weight)
    z = q.z
    c = q.c
    @inbounds begin
        xi = z[i,1]; yi = z[i,2]
        xj = z[j,1]; yj = z[j,2]
        Δx = xj-xi
        Δy = yj-yi
        p = (weightSum[i]-weightSum[j])*(Δx*(xj+xi)+Δy*(yj+yi)) - 2((c[i,1]-c[j,1])*Δx + (c[i,2]-c[j,2])*Δy)
        if i≤m && j≤m
            p += 2weight[i,j]*(Δx^2 + Δy^2)
        end
    end
    fast_int32(p)
end

# Swap encodings i and j.  Caller must provide delta(q,i,j)
function swap!( q::Quad, i::Integer, j::Integer, delta::Integer )
    w = weight
    (m,n) = size(w)
    q.d += delta
    c = q.c
    z = q.z
    xi = z[i,1]; yi = z[i,2]
    xj = z[j,1]; yj = z[j,2]
    z[i,1] = xj; z[i,2] = yj
    z[j,1] = xi; z[j,2] = yi
    # Update moments
    Δx = xj-xi
    Δy = yj-yi
    @inbounds @simd for k=1:m
        Δw = w[k,i]-w[k,j]
        c[k,1] += Δw*Δx
        c[k,2] += Δw*Δy
    end
end

# Return Quad for random mxn matrix
function randQuad( m::Integer=N, n::Integer=N )
    quad( reshape(Base.randperm(m*n),m,n) )
end

# Check that Quad functions defined so far work.
function checkQuad()
    q = randQuad()
    r = copy(q)
    @assert q==r
    for t=1:100
        n = length(q)
        i = rand(1:n)
        j = rand(1:n)
        Δq = delta(q,i,j)
        swap!(q,i,j,Δq)
        verify(q)
    end
end

==( q::Quad, r::Quad ) = q.d==r.d && q.z==r.z

# True if q and r match modulo null elements
function equiv( q::Quad, r::Quad )
    (m,n) = size(weight)
    q.d==r.d && q.z[1:m,:]==r.z[1:m,:]
end

# Construct Quad from matrix of integers
function quad{T<:Integer}( a::Matrix{T} )
    (m,n) = size(a)
    m==n==N || error("matrix must be NxN")
    z = zeros(QType,m*n,2)
    for j=1:n, i=1:m
        k = encode[a[i,j]]
        z[k,1]==0 && z[k,2]==0 || error("duplicate $(a[i,j])")
        z[k,1] = j
        z[k,2] = i
    end
    Quad(z)
end

# Optimization directions
const up = Forward
const down = Reverse

# Read Delacorte array and its Delacorte number from file written by writeQuad
function readArray( filename::String )
    open(filename,"r") do f
        s = readline(f)
        m = match(r"Size = ([0-9]+) x ([0-9])+ *",s)
        n = int(m.captures[1])
        s = readline(f)
        m = match(r"Delacorte# = ([0-9]+) *",s)
        d = int(m.captures[1])
        r = zeros(Int,n,n)
        for i=1:n
            s = readline(f)
            m = match(r"\(([0-9, ]+)\),? *",s)
            r[i,:] = map( int, split(m.captures[1],","))
        end
        (r,d)
    end
end

# Get name of file for best result known for nxn problem in direction o
function resultFileName( n::Integer, o::Ordering )
    @assert o==up || o==down
    which = o==up ? "max" : "min"
    "results/$(n)_$(which).txt"
end

# Write results/* records for problems listed in "sizes" to a file azpc.txt
# write Delacorte numbers to axpc.csv
function ship{T<:Integer}( sizes::Vector{T} )
    open("azpc.csv","w") do g
        open("azpc.txt","w") do f
            isFirst=true
            for n in sizes
                print( g, n )
                for o in Ordering[down,up]
                    (a,d) = readArray(resultFileName(n,o))
                    @assert d==slowDelacorte(a)
                    if isFirst
                        isFirst = false
                    else
                        println( f, ";\n" )
                    end
                    printAZPC( f, a )
                    print( g, ",", d )
                end
                print(g,"\n")
            end
        end
    end
end

# Read Quad from a file
function readQuad( filename::String )
    open(filename,"r") do f
        (r,d) = readArray(filename)
        q = quad(r)
        delacorte(q)==d || error("$filename has wrong Delacorte#")
        q
    end
end

# Read current best result with respect to given ordering
readBest( o::Ordering ) = readQuad(resultFileName(N,o))

readMin() = readBest(down)
readMax() = readBest(up)

# Write Quad to a file
function writeQuad( filename::String, q::Quad )
    # Verify q before opening file, to avoid corrupting file if q is bad.
    verify(q)
    open(filename,"w") do f
        show(f,q)
    end
end

# Save q if it improves on currently recorded answer.
function saveIfBetter( q::Quad, o::Ordering )
    filename = resultFileName(N,o)
    dq = delacorte(q)
    p = readQuad(filename)
    dp = delacorte(p)
    if lt(o,dp,dq)
        writeQuad(filename,q)
        print_with_color(:green,@sprintf("\aChanged %dx%d %s from %d to %d (improvement of %d)\n", N, N, o==up?"max":"min", dp, dq, dq-dp ))
    end
end

# Save q if it improves on current best answers.
function saveIfBetter( q::Quad )
    saveIfBetter(q,down)
    saveIfBetter(q,up)
end

# Get corners from a Quad
function corners( q::Quad )
    r = rect(q)
    Int[r[  1,1] r[  1,end]
        r[end,1] r[end,end]]
end

# Pin corner values from "max" solution if requested
if pinned=="corners"
    pinned = corners(readMax())
end

# ---------- indexPairArray ---------

# Pair of numbers, represented by their index encodings.
# The actual pair of values is (decode[i],decode[j]).
immutable IndexPair
    i :: Int16
    j :: Int16
end

# Create array of IndexPair of all pairs of numbers that might have non-zero
# delta when swapped.  First index is always less than second index.
function makeIndexPairArray()
    m = size(weight,1)
    n = N*N
    @assert m<n
    p = IndexPair[]
    pset = setdiff(1:n,map(i->encode[i],pinned))
    for i=1:m
        if i in pset
            for j = i+1:n
                if j in pset
                    push!(p,IndexPair(i,j))
                end
            end
        end
    end
    p
end

const indexPairArray = makeIndexPairArray()

const randIndexPairStream = RandStream(length(indexPairArray))

getRandomPair() = indexPairArray[rand(randIndexPairStream)+1]

println("indexPairArray built")

# Benchmark for measuring performance of getRandomPair
function benchmarkRandomPair()
    numIter = 10000000
    t = @elapsed for i=1:numIter; getRandomPair(); end
    @printf("%.2f MPair/sec", numIter/t/10^6)
end

# ---------- Generic optimization ----------

# Try numIter random swaps.  Accept those with delta no worse than temp
function hop!( r::Delacorte, o::Ordering, temp::Integer, numIter::Integer )
    @assert o==up || o==down
    threshold = o==up ? -temp-1 : temp+1
    for trial=1:numIter
        p = getRandomPair()
        d = delta(r,p.i,p.j)
        if lt(o,threshold,d)
            swap!(r,p.i,p.j,d)
        end
    end
end

# A temperature schedule
abstract TempSchedule

# Hyperbolic temperature schedule
immutable HyperTemp <: TempSchedule
    cutoff::Int
    val::Int
end

# Return number of iterations to run at temperature t for given schedule h
duration( h::HyperTemp, t::Integer ) = t≤h.cutoff ? h.val : round(Int,h.cutoff*h.val/t)

# Do hop! over downward range of temperatures.
function cool!( r::Delacorte, o::Ordering, startTemp::Integer, s::TempSchedule )
    for t in startTemp:-1:0
        hop!(r,o,t,duration(s,t))
    end
end

immutable SplatItem
    r::Delacorte    # Delacorte object
    temp::Integer   # Initial temperature from which it cooled
end

isless( q::SplatItem, r::SplatItem ) = q.r<r.r

# Return cur with any non-trivial prefix shared with prev replaced by "_".
function contract( prev::String, cur::String )
    i = 1
    if length(prev)==length(cur)
        while length(prev)>=i && prev[i]==cur[i]
            i += 1
        end
    end
    i>2 ?  "_"*cur[i:end] : cur
end

function string( items::Vector{SplatItem} )
    s = "["
    prev = ""
    for i in 1:length(items)
        if prev!=""
            s *= " "
        end
        cur = string(delacorte(items[i].r))
        s *= contract(prev,cur)
        prev = cur
    end
    s *= "]"
    s
end

# Remove duplicate items, assuming that duplicates are adjacent
function unique!( items::Vector{SplatItem} )
    i = 0
    for j=1:length(items)
        if i==0 || !equiv(items[j].r,items[i].r)
            i += 1
            items[i] = items[j]
        end
    end
    resize!( items, i )
end

function splat!{T<:Integer}( outItems::Vector{SplatItem}, inItems::Vector{SplatItem}, o::Ordering, start::UnitRange{T}, s::TempSchedule, arity::Integer, verbose::Bool=false )
    !verbose || @printf("splat(%s, %s, %s, %s, %d)\n", string(inItems), o==up?"up":"down", start, s, arity )
    @assert arity>=2 || first(start)==last(start)
    slope = (last(start)-first(start))/max(arity-1,1)
    for trial=0:arity-1
        for i in inItems
            r = copy(i.r)
            t = round(Int,first(start)+slope*trial)
            cool!(r,o,t,s)
            # Record temperature only if it improved the individual
            push!(outItems, SplatItem(r, lt(o,delacorte(i.r),delacorte(r)) ? t : -1))
        end
    end
    sort!( outItems, alg=MergeSort, rev=(o==up) )
    unique!( outItems )
    resize!( outItems, min(length(inItems), length(outItems)) )
    saveIfBetter( first(outItems).r, o )
end

function robo( o::Ordering, r=readBest(o); first::(Integer,Integer,Integer)=roboFirst, rest::(Integer,Integer,Integer)=roboRest, popSize::Integer=roboPop, alpha::FloatingPoint=0.7 )
    @assert 0 < alpha < 1
    inr = fill(SplatItem(r,-1),popSize)
    outr = SplatItem[]
    splat!( outr, inr, o, first[1]:first[1], HyperTemp(100,first[2]), first[3], true )
    u = round(Int,rest[1]/sqrt(alpha))
    while u>0
        @assert length(outr)==popSize
        outr, inr = inr, outr
        resize!(outr,popSize)
        for j=1:popSize
            outr[j] = SplatItem(inr[j].r,-1)
        end
        l = u≤30 ? 0 : round(Int,u*alpha)
        splat!( outr, inr, o, l:u, HyperTemp(100,rest[2]), rest[3], true )
        maxt = maximum(map(i->i.temp,outr))
        if maxt>=0
            println("    high temp=",maxt)
            u = int(maxt/sqrt(alpha))
        else
            u = l
        end
        flush(STDOUT)
    end
    global bg
    bg = outr
    nothing
end

if !isdefined(:roboFirst)
    if true
        # Default values for production runs
        roboFirst = (20000,20000,20)
        roboRest = (2000,40000,20)
        roboPop = 16
    else
        # Default Values for debugging runs
        roboFirst = (1000,2500,2)
        roboRest = (200,10000,3)
        roboPop = 4
    end
end

function flog( o::Ordering, trials::Integer )
    println("roboFirst=",roboFirst,"; roboRest=",roboRest,"; roboPop=",roboPop)
    println("pinned=",pinned)
    for t=1:trials
        println("Trial ",t," for N=",N)
        robo(o)
        best = bg[1].r
        writeQuad("trials/$(N)_$(best.d).txt",best)
        println()
    end
end

# ---------- Benchmarking ---------

function iterateCool( r::Delacorte, o::Ordering, startTemp::Integer, s::TempSchedule, trials::Integer)
    for i=1:trials
        q = copy(r)
        cool!(q,up,startTemp,s)
    end
end

function benchmarkCool( r::Delacorte )
    # Do timing run
    trials = 10
    t = @elapsed iterateCool(r,up,2000,HyperTemp(100,40000),trials)
    @printf("%f sec for N=%d\n", t/trials, N)
end

function benchmarkCool()
    # Verify that operations work
    checkQuad()
    # Warm up system (force compilation)
    cool!(randQuad(),up,10,HyperTemp(100,10))
    # Use current "max" result to start benchmark
    benchmarkCool( readMax() )
end

nothing
