#=

Program for finding extreme Delacorte Numbers
Copyright (c) 2015, Arch D. Robison
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors
   may be used to endorse or promote products derived from this software without
   specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
OF THE POSSIBILITY OF SUCH DAMAGE.

=#

import Base.match

# Swap values i and j.  Typically used for hand-editing a Delacorte rectangle.
swapVal!( r::Delacorte, i::Integer, j::Integer ) = swap!(r,encode[i],encode[j],delta(r,encode[i],encode[j]))

# ---------- Pessimization utility ---------

# Adjust r with single swap that changes delacorte(r) by amount close to desiredChange
function deopt(r::Delacorte,desiredChange)
    q = nothing
    dq = typemax(Int)
    for p in indexPairArray
        d = delta(r,p.i,p.j)
        if abs(d-desiredChange)<abs(dq-desiredChange)
            q = p
            dq = d
        end
    end
    swap!(r,q.i,q.j,dq)
    dq
end

# ---------- Inspection utilities ---------

# Return all deltas, sorted in increasing order
function swapSpectrum( r::Delacorte )
    d = Int32[delta(r,p.i,p.j) for p in indexPairArray]
    sort!(d)
end

# Return matrix of minimum deltas to move each element
function tmap( r::Delacorte, o::Ordering ) 
    (m,n) = size(weight)
    minT = fill(typemax(Int), n)
    minT[m+1:n] = 0
    factor = o==up ? -1 : 1;
    for p in indexPairArray
        t = factor*delta(r,p.i,p.j) 
        if t<minT[p.i]
            minT[p.i] = t
        end
        if t<minT[p.j]
            minT[p.j] = t
        end
    end
    rect(k->minT[k]==typemax(Int) ? -1 : minT[k],r)
end

# ---------- Matching utilities ---------

function flip!( b::Matrix, a::Matrix, which::Integer )
    (m,n)=size(a)
    @assert m==n
    @assert (m,n)==size(b)
    r = (which&2)!=0 ? reverse(1:m) : 1:1:m
    s = (which&4)!=0 ? reverse(1:n) : 1:1:n
    if (which&1)!=0
        transpose!(b,a[r,s])
    else
        b[:,:] = a[r,s]
    end
    nothing
end

function flip( a::Matrix, which::Integer )
    b = similar(a,eltype(a),size(a)...)
    flip!(b,a,which)
    b
end

# Convert zero-weight elements to 1
function strip{T<:Integer}( a::Matrix{T} )
    (m,n)=size(a)
    (wm,wn)=size(weight)
    map(a) do k
        weightSum[encode[k]] == 0 ? 1 : k
    end
end

function match( a::Matrix, b::Matrix )
    # The squaring is hack to give high weightSum extra oomph
    wa = map(i->weightSum[encode[i]]^2,a)
    sa = strip(a)
    sb = strip(b)
    sc = copy(sb)
    sw = -1
    w = -1
    for which=1:8
        flip!(sc,sb,which)
        sm = sum(wa.*int(sc.==sa))
        if sm>sw
            w = which
            sw = sm
        end
    end 
    flip!(sc,sb,w)
    a.*int(sc.==sa)
end

match( q::Quad, r::Quad ) = match(rect(q),rect(r))

function match( a::Quad, b::Quad, c::Quad... )
    x = match(a,b)
    y = match(a,c...)
    map( (i,j)->i==j ? i : 0, x, y )
end

function match( a::Vector{Quad} ) 
    length(a)>0 || error("array must be non-empty")
    match(a...)
end

function floodFill!{T<:Integer}( b::Matrix{Int}, a::Matrix{T}, i::Integer, j::Integer, Δi::Integer, Δj::Integer )
    if i in 1:size(b,1) && j in 1:size(b,2) && b[i,j]==-1
        if a[i,j]==0 || weightSum[encode[a[i,j]]]==0
            b[i,j] = 0
        else
            b[i,j] = a[i,j]
            floodFill!( b, a, i+Δi, j, Δi, Δj )
            floodFill!( b, a, i, j+Δj, Δi, Δj )
            #floodFill!( b, a, i+Δi, j+Δj, Δi, Δj )
        end
    end
end

function cornerRegions{T<:Integer}( a::Matrix{T} )
    (m,n) = size(a)
    b = fill(-1,m,n)
    floodFill!(b,a,1,1,1,1)
    floodFill!(b,a,m,1,-1,1)
    floodFill!(b,a,1,n,1,-1)
    floodFill!(b,a,m,n,-1,-1)
    map(i->i>0?i:0,b)
end

pinnedFromMatch{T<:Integer}( a::Matrix{T} ) = sort(filter(i->i!=0 && weightSum[encode[i]]!=0,a))

function pinnedFromCorners( q::Quad )
    r = rect(q)
    Int[r[1,1],r[end,1],r[1,end],r[end,end]]
end

function readTrials()
    q = Quad[]
    pattern = Regex("^$(N)_.*txt")
    for filename in readdir("trials")
        if ismatch(pattern,filename)
            push!(q,readQuad("trials/$filename"))
        end
    end
    sort!(q)
end

nothing
